package net.robowiki.knn.util;

import net.robowiki.knn.util.KNNPoint;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * 
 */
public class TestResult implements Comparable<TestResult>, Cloneable {
    // Cumulative
    private long searchTime;
    private int searchCount;
    private long worstSearchTime;
    private long addTime;
    private int addCount;
    private int repCount;
    private ArrayList<Long> searchTimes;
    // Non-cumulative
    private final int dataSize;
    private final int numNeighbours;
    private final String algorithm;
    private Double accuracy;
    private ArrayList<KNNPoint[]> results;
    private final KNNPoint[][] solution;

    public static TestResult collectResults(TestResult accumulator, TestResult newResult) {
        if (accumulator == null) {
            try {
                return (TestResult) newResult.clone();
            } catch (CloneNotSupportedException e) {
                return null; // Shouldn't happens
            }
        }
        accumulator.accumulateTimes(newResult);
        return accumulator;
    }

    public TestResult(String algorithm, int dataSize, int numNeighbours, KNNPoint[][] solution) {
        this.algorithm = algorithm;
        this.results = new ArrayList<KNNPoint[]>();
        this.searchTimes = new ArrayList<Long>();
        this.dataSize = dataSize;
        this.numNeighbours = numNeighbours;
        this.solution = solution;
        repCount = 1;
        accuracy = null;
    }

    @Override
    public Object clone() throws CloneNotSupportedException {
        TestResult result = (TestResult) super.clone();
        result.searchTimes = new ArrayList<Long>(searchTimes);
        result.results = new ArrayList<KNNPoint[]>(results);
        return result;
    }

    public void recordSearch(KNNPoint[] result, long time) {
        results.add(result);
        searchTimes.add(time);
        searchCount++;
        searchTime += time;
        worstSearchTime = Math.max(time, worstSearchTime);
    }

    public void recordAdd(long time) {
        addCount++;
        addTime += time;
    }

    private void checkAnswer() {
        if (solution == null) {
            accuracy = 1d;
            return;
        }

        int total = 0;
        int count = 0;
        for (int i = 0; i < solution.length; i++) {
            KNNPoint[] thisSolution = solution[i];
            total += thisSolution.length;
            if (i >= results.size()) {
                continue;
            }

            String[] solutionStrings = new String[thisSolution.length];
            double largestSol = 0;
            for (int j = 0; j < thisSolution.length; j++) {
                solutionStrings[j] = thisSolution[j].getValue();
                largestSol = Math.max(thisSolution[j].getDistance(), largestSol);
            }

            KNNPoint[] thisResult = results.get(i);
            double largestdist = 0;
            String[] resultStrings = new String[thisResult.length];
            for (int j = 0; j < thisResult.length; j++) {
                resultStrings[j] = thisResult[j].getValue();
                largestdist = Math.max(thisResult[j].getDistance(), largestdist);
            }
            List<String> abc = Arrays.asList(solutionStrings);

            for (String s : resultStrings) {
                if (abc.contains(s))
                    count++;
            }
        }

        accuracy = ((double) count / total);
    }


    private void accumulateTimes(TestResult other) {
        for (int i = 0; i < searchTimes.size(); i++) {
            this.searchTimes.set(i, this.searchTimes.get(i) + other.searchTimes.get(i));
        }
        this.searchTime += other.searchTime;
        this.searchCount += other.searchCount;
        this.worstSearchTime += other.worstSearchTime;
        this.addTime += other.addTime;
        this.addCount += other.addCount;
        this.repCount += other.repCount;
    }

    public String toString() {
        StringBuilder sb = new StringBuilder();
        DecimalFormat time_df = new DecimalFormat("0.0000");
        DecimalFormat percent_df = new DecimalFormat("0.0");

        sb.append("RESULT << k-nearest neighbours search with ").append(
                getAlgorithmName()).append(" >>").append("\n");
        sb.append(": Average searching time       = ").append(
                time_df.format(getAverageSearchTime() * 1E3)).append(" miliseconds\n");
        sb.append(": Average worst searching time = ").append(
                time_df.format(getAverageWorstSearchTime() * 1E3)).append(" miliseconds\n");
        sb.append(": Average adding time          = ").append(
                time_df.format(getAverageAddTime() * 1E6)).append(" microseconds\n");
        sb.append(": Accuracy                     = ").append(
                percent_df.format(getAccuracy() * 100)).append("%\n");

        return sb.toString();
    }

    public int compareTo(TestResult arg0) {
        return Double.compare(getAverageSearchTime(), arg0.getAverageSearchTime());
    }

    public double getAverageSearchTime() {
        return (((double)searchTime) / searchCount)/1E9;
    }

    public double getAverageWorstSearchTime() {
        return (((double)worstSearchTime) / repCount)/1E9;
    }

    public double getAverageAddTime() {
        return (((double)addTime) / addCount)/1E9;
    }

    public double getAccuracy() {
        if (accuracy == null) checkAnswer();
        return accuracy;
    }

    public String getAlgorithmName() {
        return algorithm;
    }

    public KNNPoint[][] getSearchResults() {
        return results.toArray(new KNNPoint[results.size()][]);
    }

    public double[] getSearchTimes() {
        double[] searchTimesArray = new double[searchTimes.size()];
        for (int i = 0; i < searchTimesArray.length; i++) {
            searchTimesArray[i] = (((double)searchTimes.get(i))/repCount)/1E9;
        }
        return searchTimesArray;
    }
}
